package adapter

import (
	"context"
	"database/sql"
	"errors"
	"fmt"
	"time"

	"gitea.dwysokinski.me/twhelp/corev3/internal/bun/bunmodel"
	"gitea.dwysokinski.me/twhelp/corev3/internal/domain"
	"github.com/uptrace/bun"
	"github.com/uptrace/bun/dialect"
)

type TribeBunRepository struct {
	db bun.IDB
}

func NewTribeBunRepository(db bun.IDB) *TribeBunRepository {
	return &TribeBunRepository{db: db}
}

func (repo *TribeBunRepository) CreateOrUpdate(ctx context.Context, params ...domain.CreateTribeParams) error {
	if len(params) == 0 {
		return nil
	}

	now := time.Now()
	tribes := make(bunmodel.Tribes, 0, len(params))

	for _, p := range params {
		base := p.Base()
		tribes = append(tribes, bunmodel.Tribe{
			ID:                base.ID(),
			ServerKey:         p.ServerKey(),
			Name:              base.Name(),
			Tag:               base.Tag(),
			NumMembers:        base.NumMembers(),
			NumVillages:       base.NumVillages(),
			Points:            base.Points(),
			AllPoints:         base.AllPoints(),
			Rank:              base.Rank(),
			ProfileURL:        base.ProfileURL().String(),
			BestRank:          p.BestRank(),
			BestRankAt:        p.BestRankAt(),
			MostPoints:        p.MostPoints(),
			MostPointsAt:      p.MostPointsAt(),
			MostVillages:      p.MostVillages(),
			MostVillagesAt:    p.MostVillagesAt(),
			CreatedAt:         now,
			OpponentsDefeated: bunmodel.NewOpponentsDefeated(base.OD()),
		})
	}

	q := repo.db.NewInsert().
		Model(&tribes)

	//nolint:exhaustive
	switch q.Dialect().Name() {
	case dialect.PG:
		q = q.On("CONFLICT ON CONSTRAINT tribes_pkey DO UPDATE")
	case dialect.SQLite:
		q = q.On("CONFLICT(id, server_key) DO UPDATE")
	default:
		q = q.Err(errors.New("unsupported dialect"))
	}

	if _, err := q.
		Set("name = EXCLUDED.name").
		Set("tag = EXCLUDED.tag").
		Set("num_members = EXCLUDED.num_members").
		Set("num_villages = EXCLUDED.num_villages").
		Set("points = EXCLUDED.points").
		Set("all_points = EXCLUDED.all_points").
		Set("rank = EXCLUDED.rank").
		Set("profile_url = EXCLUDED.profile_url").
		Set("best_rank = EXCLUDED.best_rank").
		Set("best_rank_at = EXCLUDED.best_rank_at").
		Set("most_villages = EXCLUDED.most_villages").
		Set("most_villages_at = EXCLUDED.most_villages_at").
		Set("most_points = EXCLUDED.most_points").
		Set("most_points_at = EXCLUDED.most_points_at").
		Set("deleted_at = EXCLUDED.deleted_at").
		Apply(appendODSetClauses).
		Returning("").
		Exec(ctx); err != nil {
		return fmt.Errorf("something went wrong while inserting tribes into the db: %w", err)
	}

	return nil
}

func (repo *TribeBunRepository) UpdateDominance(ctx context.Context, serverKey string, numPlayerVillages int) error {
	q := repo.db.NewUpdate().
		Model((*bunmodel.Tribe)(nil)).
		Returning("NULL").
		Where("server_key = ?", serverKey).
		Where("deleted_at IS NULL")

	if numPlayerVillages > 0 {
		q = q.Set("dominance = CAST(num_villages as double precision) / ? * 100", numPlayerVillages)
	} else {
		q = q.Set("dominance = 0")
	}

	if _, err := q.Exec(ctx); err != nil {
		return fmt.Errorf("%s: couldn't update dominance: %w", serverKey, err)
	}

	return nil
}

func (repo *TribeBunRepository) List(ctx context.Context, params domain.ListTribesParams) (domain.Tribes, error) {
	var tribes bunmodel.Tribes

	if err := repo.db.NewSelect().
		Model(&tribes).
		Apply(listTribesParamsApplier{params: params}.apply).
		Scan(ctx); err != nil && !errors.Is(err, sql.ErrNoRows) {
		return nil, fmt.Errorf("couldn't select tribes from the db: %w", err)
	}

	return tribes.ToDomain()
}

func (repo *TribeBunRepository) Delete(ctx context.Context, serverKey string, ids ...int) error {
	if len(ids) == 0 {
		return nil
	}

	if _, err := repo.db.NewUpdate().
		Model((*bunmodel.Tribe)(nil)).
		Where("deleted_at IS NULL").
		Where("id IN (?)", bun.In(ids)).
		Where("server_key = ?", serverKey).
		Set("deleted_at = ?", time.Now()).
		Returning("").
		Exec(ctx); err != nil {
		return fmt.Errorf("couldn't delete tribes: %w", err)
	}

	return nil
}

type listTribesParamsApplier struct {
	params domain.ListTribesParams
}

//nolint:gocyclo
func (a listTribesParamsApplier) apply(q *bun.SelectQuery) *bun.SelectQuery {
	if ids := a.params.IDs(); len(ids) > 0 {
		q = q.Where("tribe.id IN (?)", bun.In(ids))
	}

	if idGT := a.params.IDGT(); idGT.Valid {
		q = q.Where("tribe.id > ?", idGT.Value)
	}

	if serverKeys := a.params.ServerKeys(); len(serverKeys) > 0 {
		q = q.Where("tribe.server_key IN (?)", bun.In(serverKeys))
	}

	if deleted := a.params.Deleted(); deleted.Valid {
		if deleted.Value {
			q = q.Where("tribe.deleted_at IS NOT NULL")
		} else {
			q = q.Where("tribe.deleted_at IS NULL")
		}
	}

	for _, s := range a.params.Sort() {
		switch s {
		case domain.TribeSortIDASC:
			q = q.Order("tribe.id ASC")
		case domain.TribeSortIDDESC:
			q = q.Order("tribe.id DESC")
		case domain.TribeSortServerKeyASC:
			q = q.Order("tribe.server_key ASC")
		case domain.TribeSortServerKeyDESC:
			q = q.Order("tribe.server_key DESC")
		default:
			return q.Err(errors.New("unsupported sort value"))
		}
	}

	return q.Limit(a.params.Limit()).Offset(a.params.Offset())
}
